#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from io import BytesIO

import unittest
import unittest.mock

from i75.image import SingleColourImage


class TestImage(unittest.TestCase):
    def test_single_color_image_is_pixel(self):
        image_data = BytesIO((1 << 6).to_bytes(1, byteorder="big")
                             + (1 << 7).to_bytes(1, byteorder="big"))
        sci = SingleColourImage(2, 2, image_data)

        self.assertFalse(sci._is_pixel(0, 0))
        self.assertTrue(sci._is_pixel(1, 0))
        self.assertTrue(sci._is_pixel(0, 1))
        self.assertFalse(sci._is_pixel(1, 1))



# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
try:
    from typing import Tuple, Optional
except ImportError:
    pass

from .graphics import Graphics


class Image:
    data: bytearray

    def __init__(self, width: int, height: int) -> None:
        self.width = width
        self.height = height

    def render(self, buffer: Graphics, offset_x: int, offset_y: int) -> None:
        raise NotImplementedError()

    def set_colour(self, red: int, green: int, blue: int) -> None:
        raise NotImplementedError()

    @staticmethod
    def load(fp) -> "Image":
        magic = fp.read(5)
        if magic != b"I75v1":
            raise ValueError(
                "Image has the wrong initial bytes. Wrong format?")
        width = int.from_bytes(fp.read(1), "big")
        height = int.from_bytes(fp.read(1), "big")
        colours = int.from_bytes(fp.read(1), "big")

        if colours == 1:
            return SingleColourImage(width, height, fp)
        if colours == 3:
            return ThreeColourImage(width, height, fp)
        raise ValueError("Image has an unsupported number of colours.")

    @staticmethod
    def load_into_buffer(fp, data: bytearray) -> "Image":
        magic = fp.read(5)
        if magic != b"I75v1":
            raise ValueError(
                "Image has the wrong initial bytes. Wrong format?")
        width = int.from_bytes(fp.read(1), "big")
        height = int.from_bytes(fp.read(1), "big")
        colours = int.from_bytes(fp.read(1), "big")

        if colours == 1:
            assert len(data) >= width * math.ceil(width / 8.0)
            return SingleColourImage(width, height, fp, data)
        if colours == 3:
            assert len(data) >= width * height * 3
            return ThreeColourImage(width, height, fp, data)
        raise ValueError("Image has an unsupported number of colours.")


class SingleColourImage(Image):
    def __init__(self,
                 width: int,
                 height: int,
                 fp,
                 data: Optional[bytearray] = None) -> None:
        super().__init__(width, height)
        if data is None:
            self.data: bytearray = fp.read(height * math.ceil(width / 8.0))
        else:
            self.data = data
            for i in range(0, height * math.ceil(width / 8.0)):
                self.data[i] = ord(fp.read(1))
        self.colour: Tuple[int, int, int] = (255, 255, 255)

    def set_colour(self, red: int, green: int, blue: int) -> None:
        self.colour = (red, green, blue)

    def render(self, buffer: Graphics, offset_x: int, offset_y: int) -> None:
        buffer.set_pen(buffer.create_pen(*self.colour))
        for y in range(self.height):
            for x in range(self.width):
                if self._is_pixel(x, y):
                    buffer.pixel(offset_x + x, offset_y + y)

    def _is_pixel(self, x: int, y: int) -> bool:
        row_width = math.ceil(self.width / 8.0)
        byte = self.data[y * row_width + math.floor(x / 8)]
        return (byte >> (7 - x % 8)) & 1 == 1


class ThreeColourImage(Image):
    def __init__(self,
                 width: int,
                 height: int,
                 fp,
                 data: Optional[bytearray] = None) -> None:
        super().__init__(width, height)
        if data is None:
            self.data: bytearray = fp.read(3 * width * height)
        else:
            self.data = data
            for i in range(0, 3 * width * height):
                self.data[i] = ord(fp.read(1))

    def render(self, buffer: Graphics, offset_x: int, offset_y: int) -> None:
        for y in range(self.height):
            for x in range(self.width):
                buffer.set_pen(buffer.create_pen(
                    self.data[3 * (y * self.width + x)],
                    self.data[3 * (y * self.width + x) + 1],
                    self.data[3 * (y * self.width + x) + 2],
                ))
                buffer.pixel(offset_x + x, offset_y + y)



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import unittest.mock

from i75.text import render_text, text_boundingbox

FONT = "cg_pixel_3x5_5"


class TestText(unittest.TestCase):
    def test_bounding_box_width_matches_render(self) -> None:
        text = "This Is A Test String."
        width, _ = text_boundingbox(FONT, text)

        mock_buffer = unittest.mock.Mock()
        render_text(mock_buffer, FONT, 0, 0, text)

        max_x = max([x for ((x, _), _) in mock_buffer.pixel.call_args_list])

        self.assertEqual(width, max_x + 2)



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
try:
    from typing import Dict, Optional
except ImportError:
    pass


def get(url: str,
        headers: Optional[Dict[str, str]] = None,
        data: Optional[str] = None,
        stream: bool = False) -> requests.Response:
    return requests.get(url, headers=headers, data=data, stream=stream)


def post(url: str,
         headers: Optional[Dict[str, str]] = None,
         data: Optional[str] = None,
         stream: bool = False) -> requests.Response:
    return requests.post(url, headers=headers, data=data, stream=stream)



#!/usr/bin/env micropython
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Tuple


class Pen:
    def __init__(self, r: int, g: int, b: int, a: int = 255) -> None:
        self._value = r << 24 | g << 16 | b << 8 | a

    @property
    def r(self) -> int:
        return (self._value >> 24) & 255

    @property
    def g(self) -> int:
        return (self._value >> 16) & 255

    @property
    def b(self) -> int:
        return (self._value >> 8) & 255

    @property
    def a(self) -> int:
        return self._value & 255

    def as_tuple(self) -> Tuple[int, int, int]:
        return (self.r, self.g, self.b)


class RGB332Pen(Pen):
    def __init__(self, r: int, g: int, b: int) -> None:
        super().__init__((1 << 3) * round(float(r) / (1 << 3)),
                         (1 << 3) * round(float(g) / (1 << 3)),
                         (1 << 2) * round(float(b) / (1 << 2)))


class RGB888Pen(Pen):
    def __init__(self, r: int, g: int, b: int) -> None:
        super().__init__(r, g, b)



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import unittest.mock

from i75 import DateTime
from i75.tz import EuropeLondon


class TestEuropeLondon(unittest.TestCase):
    def test_to_local_before_clock_change(self):
        dt = DateTime(2023, 1, 1)

        self.assertEqual(dt, EuropeLondon.to_localtime(dt))

    def test_to_local_after_clock_change_back(self):
        dt = DateTime(2023, 11, 1)

        self.assertEqual(dt, EuropeLondon.to_localtime(dt))

    def test_to_local_after_clock_change(self):
        dt = DateTime(2023, 4, 1)

        self.assertEqual(DateTime(2023, 4, 1, hour=1),
                         EuropeLondon.to_localtime(dt))

    def test_to_local_march_before(self):
        dt = DateTime(2023, 3, 1)

        self.assertEqual(dt, EuropeLondon.to_localtime(dt))

    def test_to_local_march_after(self):
        dt = DateTime(2023, 3, 29)

        self.assertEqual(DateTime(2023, 3, 29, hour=1),
                         EuropeLondon.to_localtime(dt))

    def test_to_local_march_on_day(self):
        dt = DateTime(2023, 3, 26, hour=2)

        self.assertEqual(DateTime(2023, 3, 26, hour=3),
                         EuropeLondon.to_localtime(dt))

    def test_to_local_october_before(self):
        dt = DateTime(2023, 10, 1)

        self.assertEqual(DateTime(2023, 10, 1, hour=1),
                         EuropeLondon.to_localtime(dt))

    def test_to_local_october_after(self):
        dt = DateTime(2023, 10, 30)

        self.assertEqual(dt, EuropeLondon.to_localtime(dt))

    def test_to_local_october_on_day(self):
        dt = DateTime(2023, 10, 29, hour=3)

        self.assertEqual(dt, EuropeLondon.to_localtime(dt))

    def test_to_utc_before_clock_change(self):
        dt = DateTime(2023, 1, 1)

        self.assertEqual(dt, EuropeLondon.to_utc(dt))

    def test_to_utc_after_clock_change_back(self):
        dt = DateTime(2023, 11, 1)

        self.assertEqual(dt, EuropeLondon.to_utc(dt))

    def test_to_utc_after_clock_change(self):
        dt = DateTime(2023, 4, 1, hour=1)

        self.assertEqual(DateTime(2023, 4, 1), EuropeLondon.to_utc(dt))

    def test_to_utc_march_before(self):
        dt = DateTime(2023, 3, 1)

        self.assertEqual(dt, EuropeLondon.to_utc(dt))

    def test_to_utc_march_after(self):
        dt = DateTime(2023, 3, 29, hour=1)

        self.assertEqual(DateTime(2023, 3, 29), EuropeLondon.to_utc(dt))

    def test_to_utc_march_on_day(self):
        dt = DateTime(2023, 3, 26, hour=3)

        self.assertEqual(DateTime(2023, 3, 26, hour=2),
                         EuropeLondon.to_utc(dt))

    def test_to_utc_october_before(self):
        dt = DateTime(2023, 10, 1, hour=1)

        self.assertEqual(DateTime(2023, 10, 1), EuropeLondon.to_utc(dt))

    def test_to_utc_october_after(self):
        dt = DateTime(2023, 10, 30)

        self.assertEqual(dt, EuropeLondon.to_utc(dt))

    def test_to_utc_october_on_day(self):
        dt = DateTime(2023, 10, 29, hour=3)

        self.assertEqual(dt, EuropeLondon.to_utc(dt))



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    from typing import Any, Callable, Dict, Tuple, cast
except ImportError:
    def cast(x, y):  # type:ignore
        return y

from .graphics import Graphics


class FontData:
    HEIGHT: int
    SPACE_WIDTH: int
    DATA: Dict[str, Tuple[int, bytes]]


def __get_font(font: str) -> FontData:
    globals: Dict[str, Any] = {}
    locals: Dict[str, Any] = {}
    exec(f"import i75.fontdata.{font}", globals, locals)
    return cast(FontData, getattr(locals["i75"].fontdata, font))


def __scale_pixel(scale: int,
                  pixel: Callable[[int, int], None]) \
        -> Callable[[int, int, int, int], None]:
    def r(x: int, y: int, cx: int, cy: int) -> None:
        for sx in range(scale):
            for sy in range(scale):
                pixel(x + cx * scale + sx, y + cy * scale + sy)
    return r


def render_text(buffer: Graphics,
                font: str,
                x: int,
                y: int,
                text: str,
                scale: int = 1) -> None:
    """
    Render the given text, at location (x,y) using font onto the scren buffer.
    """
    if "\n" in text:
        render_text_multiline(buffer, font, x, y, text)
        return

    font_data = __get_font(font)

    if scale == 1:
        pixel: Callable[[int, int, int, int], None] = \
            lambda x, y, cx, cy: buffer.pixel(x + cx, y + cy)
    else:
        pixel = __scale_pixel(scale, buffer.pixel)

    for c in text.upper():
        if c == " ":
            x += font_data.SPACE_WIDTH * scale
            continue
        if c in font_data.DATA:
            width, data = font_data.DATA[c]
        else:
            width, data = font_data.DATA["#"]

        for cy in range(font_data.HEIGHT):
            row = data[cy]
            if row == b'\0':
                continue
            for cx in range(width):
                if row & 1 == 1:
                    pixel(x, y, cx, cy)
                row = row >> 1
        x += (width + 1) * scale


def render_text_multiline(buffer: Graphics,
                          font: str,
                          x: int,
                          y: int,
                          textdata: str,
                          scale: int = 1) -> None:
    font_data = __get_font(font)

    for line in textdata.split("\n"):
        render_text(buffer, font, x, y, line, scale)

        y += font_data.HEIGHT


def text_boundingbox(font: str, text: str, scale: int = 1) -> Tuple[int, int]:
    """
    Return the width and height of the given text using the given font.
    """
    font_data = __get_font(font)

    width, height = 0, 0
    for line in text.split("\n"):
        line_width = 0
        for c in line.upper():
            if c == " ":
                line_width += font_data.SPACE_WIDTH
                continue
            if c in font_data.DATA:
                cwidth, _ = font_data.DATA[c]
            else:
                cwidth, _ = font_data.DATA["#"]
            line_width += cwidth + 1
        if line_width > width:
            width = line_width
        height += font_data.HEIGHT

    return width * scale, height * scale


def wrap_text(font: str, text: str, max_width: int, scale: int = 1) -> str:
    lines = [text]
    while True:
        split_line = False
        for i in range(len(lines)):
            width = text_boundingbox(font, lines[i], scale)[0]
            if width > max_width and " " in lines[i]:
                split_line = True
                last_word = lines[i].split(" ")[-1]
                lines[i] = " ".join(lines[i].split(" ")[:-1])
                if len(lines) == i + 1:
                    lines.append(last_word)
                else:
                    lines[i + 1] = last_word + " " + lines[i + 1]

        if not split_line:
            break

    return "\n".join(lines)



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Tuple
import unittest
import unittest.mock

import picographics

from i75.graphics import Graphics


@unittest.mock.patch("hub75.pygame")
@unittest.mock.patch("i75.graphics.picographics")
class TestLineDrawing(unittest.TestCase):
    def create(self, mockgraphics: unittest.mock.Mock) \
            -> Tuple[unittest.mock.Mock, Graphics]:
        mock = unittest.mock.Mock()

        mock.get_bounds.return_value = (64, 64)
        mockgraphics.PicoGraphics.return_value = mock

        graphics = Graphics(picographics.DISPLAY_INTERSTATE75_64X64)

        return (mock, graphics)

    def test_diag_down(self, mockgraphics: unittest.mock.Mock, _) -> None:
        mock, graphics = self.create(mockgraphics)

        graphics.line(5, 5, 0, 0)

        mock.pixel.assert_called()
        for call, expected in zip(mock.pixel.mock_calls,
                                  [(5, 5), (4, 4), (3, 3),
                                   (2, 2), (1, 1), (0, 0)]):
            self.assertEqual(call[1], expected)

    def test_diag_up(self, mockgraphics: unittest.mock.Mock, _) -> None:
        mock, graphics = self.create(mockgraphics)

        graphics.line(0, 5, 5, 0)

        mock.pixel.assert_called()
        for call, expected in zip(mock.pixel.mock_calls,
                                  [(0, 5), (1, 4), (2, 3),
                                   (3, 2), (4, 1), (5, 0)]):
            self.assertEqual(call[1], expected)

    def test_steep_up(self, mockgraphics: unittest.mock.Mock, _) -> None:
        mock, graphics = self.create(mockgraphics)

        graphics.line(0, 5, 2, 0)

        mock.pixel.assert_called()
        for call, expected in zip(mock.pixel.mock_calls,
                                  [(0, 5), (0, 4), (1, 3),
                                   (1, 2), (2, 1), (2, 0)]):
            self.assertEqual(call[1], expected)

    def test_horizontal(self, mockgraphics: unittest.mock.Mock, _) -> None:
        mock, graphics = self.create(mockgraphics)

        graphics.line(0, 0, 0, 5)

        mock.pixel.assert_called()
        for call, expected in zip(mock.pixel.mock_calls,
                                  [(0, 0), (0, 1), (0, 2),
                                   (0, 3), (0, 4), (0, 5)]):
            self.assertEqual(call[1], expected)

    def test_vertical(self, mockgraphics: unittest.mock.Mock, _) -> None:
        mock, graphics = self.create(mockgraphics)

        graphics.line(0, 0, 5, 0)

        mock.pixel.assert_called()
        for call, expected in zip(mock.pixel.mock_calls,
                                  [(0, 0), (1, 0), (2, 0),
                                   (3, 0), (4, 0), (5, 0)]):
            self.assertEqual(call[1], expected)



#!/usr/bin/env micropython
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime as PyDateTime
import math
import time
from typing import Callable, Optional

import picographics

from .basei75 import BaseI75
from .datetime import DateTime


class EmulatedI75(BaseI75):
    """
    This class aims to emulates all the features of the Interstate75, using
    PyGame to render the display. See :ref baseinterstate75: for a description
    of the available methods.
    """
    def __init__(self,
                 display_type: picographics.DisplayType,
                 stb_invert=False,
                 rotate: int = 0) -> None:
        super().__init__(display_type,
                         rotate=rotate,
                         wifi_ssid="native",
                         wifi_password="native")

    @staticmethod
    def is_emulated() -> bool:
        return True

    def set_time(self) -> bool:
        return self.wlan is not None and self.wlan.isconnected()

    def ticks_ms(self) -> int:
        return math.floor(time.time_ns() / 1000000)

    def ticks_diff(self, t1: int, t2: int) -> int:
        return t1 - t2

    def sleep_ms(self, delay: int) -> None:
        time.sleep(delay / 1000.0)

    def now(self) -> DateTime:
        now = PyDateTime.utcnow()
        return DateTime(now.year,
                        now.month,
                        now.day,
                        now.weekday(),
                        now.hour,
                        now.minute,
                        now.second,
                        now.microsecond)
